/* Data Abort handling
   Copyright (c) 2010 Networked Systems Group, University of Kaiserslautern

This file is part of the Imote2 Image Loader.

The Imote2 Image Loader is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

The Imote2 Image Loader is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with the Imote2 Image Loader. If not, see <http://www.gnu.org/licenses/>.
*/
#include "exception.h"
#include "exceptions.h"
#include "ledcycle.h"
#include "doublefault.h"
#include "backtrace.h"
#include <stdio.h>

static void printFault(uint32_t lr, const uint32_t * fp, uint32_t fault,
	uint32_t faultAddress, const struct _ExcProcessorStatus * pstat);

__attribute__((interrupt("ABORT"),weak))
void hplarmv_dabort() {
	/* gather some data: lr, fp, fault source, fault address and processor status */
	uint32_t lr = (uint32_t)__builtin_return_address(0);
	uint32_t * fp = (uint32_t*)__builtin_frame_address(0);
	uint32_t fault = _excGetFaultStatus();
	uint32_t faultAddress = _excGetFaultAddress();
	struct _ExcProcessorStatus pstat;
	_excGetProcessorStatus(&pstat);
	
	/* check for double faults */
	_excDoubleFaultCheck(lr, fp, &pstat);
	
	/* initialize exception handler */
	_excInit();
	
	/* begin in next line */
	_excUartOut("\n");
	
	while (1) {
		printFault(lr, fp, fault, faultAddress, &pstat);
		_excLedCycle(TRUE);
	}
}

static void printFault(uint32_t lr, const uint32_t * fp, uint32_t fault,
	uint32_t faultAddress, const struct _ExcProcessorStatus * pstat) {
	char buf[128];
	
	/* print the exception type and the old mode */
	snprintf(buf, sizeof(buf), "DATA ABORT occured in %s mode\n", pstat->spsrModeLabel);
	_excUartOut(buf);
	
	/* now get the abortion source and fault address */
	uint32_t src = ((fault >> 6) & 0x10) | (fault & 0xf);
	bool domVal = FALSE, farVal = FALSE;
	const char * sourceLabel = NULL;
	switch (src) {
		case 0x1:
		case 0x3:
			/* Alignment fault */
			farVal = TRUE;
			sourceLabel = "Alignment Fault";
			break;
		case 0xc:
			/* External Abort on Translation (1st level) */
			farVal = TRUE;
			sourceLabel = "External Abort on first level Translation";
			break;
		case 0xe:
			/* External Abort on Translation (2nd level) */
			domVal = TRUE;
			farVal = TRUE;
			sourceLabel = "External Abort on second level Translation";
			break;
		case 0x5:
			/* Translation (Section) */
			farVal = TRUE;
			sourceLabel = "Section Translation Fault";
			break;
		case 0x7:
			/* Translation (Page) */
			domVal = TRUE;
			farVal = TRUE;
			sourceLabel = "Page Translation Fault";
			break;
		case 0x9:
			/* Domain (Section) */
			domVal = TRUE;
			farVal = TRUE;
			sourceLabel = "Section Domain Fault";
			break;
		case 0xb:
			/* Domain (Page) */
			domVal = TRUE;
			farVal = TRUE;
			sourceLabel = "Page Domain Fault";
			break;
		case 0xd:
			/* Permission (Section) */
			domVal = TRUE;
			farVal = TRUE;
			sourceLabel = "Section Permission Fault";
			break;
		case 0xf:
			/* Permission (Page) */
			domVal = TRUE;
			farVal = TRUE;
			sourceLabel = "Page Permission Fault";
			break;
		case 0x14:
			/* Lock Abort */
			sourceLabel = "Lock Abort";
			break;
		case 0x16:
			/* Imprecise External Data Abort */
			sourceLabel = "Imprecise External Data Abort";
			break;
		case 0x18:
			/* Data Cache Parity Error Exception */
			sourceLabel = "Data Cache Parity Error";
			break;
		default:
			sourceLabel = "Unknown Abort";
			break;
	}
	/* print the pretty printed fault status */
	snprintf(buf, sizeof(buf), " Fault Status: %s\n", sourceLabel);
	_excUartOut(buf);
	/* and some more information */
	snprintf(buf, sizeof(buf), " FSR:  0x%08x  Domain: 0x%x %-9s  FAR: 0x%08x %s  SP: 0x%08x\n",
		fault, (fault >> 4) & 0xf, domVal ? _EXC_LABEL_VALID : _EXC_LABEL_INVALID,
		faultAddress, farVal ? _EXC_LABEL_VALID : _EXC_LABEL_INVALID, pstat->oldsp);
	_excUartOut(buf);
	
	_excPrintProcessorStatus(lr, fp, pstat);
	
	_excBacktrace(lr, fp);
}

